#pragma once

#include "plugin.h"
#include "sb.h"

#define DEF_STORE(NAME, TYPE, MAX_SIZE) \
    typedef struct NAME {               \
        TYPE data[MAX_SIZE];            \
        size_t count;                   \
    } NAME##_t

#define FOREACH_STORE(VALUE_NAME, STORE) \
    for (typeof(&STORE.data[0]) VALUE_NAME = STORE.data; VALUE_NAME < STORE.data + STORE.count; ++VALUE_NAME)

#define FOREACH_LIST(VALUE_NAME, LIST) \
    for (typeof(LIST) VALUE_NAME = LIST; VALUE_NAME; VALUE_NAME = VALUE_NAME->next_tied)

typedef struct disk_info {
    imsm_super_t* sb;
    imsm_disk_t* disk;
    struct disk_info* next_tied;
    storage_object_t* evms_data;
    u_int8_t index;
} disk_info_t;

DEF_STORE(disk_infos_store, disk_info_t, 1024);

typedef struct container_info {
    imsm_super_t* sb; // owner
    disk_info_t* disks;
    struct volume_info* volumes;
    storage_container_t* evms_data;
    int md_minor;
    u_int8_t active;
} container_info_t;

DEF_STORE(container_infos_store, container_info_t, 256);

typedef struct volume_info {
    imsm_dev_t* dev;
    container_info_t* container;
    struct volume_info* next_tied;
    storage_object_t* evms_data;
    u_int8_t index;
} volume_info_t;

DEF_STORE(volume_infos_store, volume_info_t, 1024);

typedef enum controller_cap {
    controller_cap_raid0 = 1 << 0,
    controller_cap_raid1 = 1 << 1,
    controller_cap_raid10 = 1 << 2,
    controller_cap_raid5 = 1 << 3,
} controller_cap_e;

static inline boolean check_level_capability(controller_cap_e cap, uint8_t level) {
    controller_cap_e l = 0;
    switch (level) {
    case 0:
        l = controller_cap_raid0;
        break;
    case 1:
        l = controller_cap_raid1;
        break;
    case 5:
        l = controller_cap_raid5;
        break;
    case 10:
        l = controller_cap_raid10;
        break;
    }
    return cap & l;
}

typedef struct controller_info {
    char sysfs_path[4096];
    controller_cap_e capabilities;
    struct controller_info* next_tied;
} controller_info_t;

#undef DEF_STORE
