#pragma once

#include <common.h>

/*--------- MD superblock IMSM format ------------*/
#define MPB_SIGNATURE "Intel Raid ISM Cfg Sig. "
#define MPB_SIG_LEN (strlen(MPB_SIGNATURE))
#define MPB_VERSION_RAID0 "1.0.00"
#define MPB_VERSION_RAID1 "1.1.00"
#define MPB_VERSION_MANY_VOLUMES_PER_ARRAY "1.2.00"
#define MPB_VERSION_3OR4_DISK_ARRAY "1.2.01"
#define MPB_VERSION_RAID5 "1.2.02"
#define MPB_VERSION_5OR6_DISK_ARRAY "1.2.04"
#define MPB_VERSION_CNG "1.2.06"
#define MPB_VERSION_ATTRIBS "1.3.00"

#define MD_IMSM_BYTES 1024
#define MD_IMSM_SECTOR_SIZE 512
#define MD_IMSM_SECTORS (MD_IMSM_BYTES / MD_IMSM_SECTOR_SIZE)

#define MAX_SIGNATURE_LENGTH 32
#define MAX_RAID_SERIAL_LEN 16
#define NUM_BLOCKS_DIRTY_STRIPE_REGION 2048

/* Disk configuration info. */
#define IMSM_MAX_DEVICES 255
typedef struct imsm_disk {
    char serial[MAX_RAID_SERIAL_LEN];            /* 0xD8 - 0xE7 ascii serial number */
    u_int32_t total_blocks_lo;                   /* 0xE8 - 0xEB total blocks lo */
    u_int32_t scsi_id;                           /* 0xEC - 0xEF scsi ID */
#define SPARE_DISK      __cpu_to_le32(0x01)      /* Spare */
#define CONFIGURED_DISK __cpu_to_le32(0x02)      /* Member of some RaidDev */
#define FAILED_DISK     __cpu_to_le32(0x04)      /* Permanent failure */
#define JOURNAL_DISK    __cpu_to_le32(0x2000000) /* Device marked as Journaling Drive */
    u_int32_t status;                            /* 0xF0 - 0xF3 */
    u_int32_t owner_cfg_num;                     /* which config 0,1,2... owns this disk */
    u_int32_t total_blocks_hi;                   /* 0xF4 - 0xF5 total blocks hi */
#define IMSM_DISK_FILLERS 3
    u_int32_t _filler[IMSM_DISK_FILLERS];        /* 0xF5 - 0x107 MPB_DISK_FILLERS for future expansion */
} imsm_disk_t;

/* map selector for map managment
 */
#define MAP_0 0
#define MAP_1 1
#define MAP_X -1

/* RAID map configuration infos. */
typedef struct imsm_map {
    u_int32_t pba_of_lba0_lo;       /* start address of partition */
    u_int32_t blocks_per_member_lo; /* blocks per member */
    u_int32_t num_data_stripes_lo;  /* number of data stripes */
    u_int16_t blocks_per_strip;
    u_int8_t map_state;             /* normal, uninitialized, degraded, failed */
#define imsm_t_state_normal        0
#define imsm_t_state_uninitialized 1
#define imsm_t_state_degraded      2
#define imsm_t_state_failed        3
    u_int8_t raid_level;
#define imsm_t_raid0 0
#define imsm_t_raid1 1
#define imsm_t_raid5 5        /* since metadata version 1.2.02 ? */
    u_int8_t num_members;     /* number of member disks */
    u_int8_t num_domains;     /* number of parity domains */
    u_int8_t failed_disk_num; /* valid only when state is degraded */
    u_int8_t ddf;
    u_int32_t pba_of_lba0_hi;
    u_int32_t blocks_per_member_hi;
    u_int32_t num_data_stripes_hi;
    u_int32_t _filler[4];      /* expansion area */
#define imsm_ord_rebuild (1 << 24)
    u_int32_t disk_ord_tbl[1]; /* disk_ord_tbl[num_members],
                                * top byte contains some flags
                                */
} imsm_map_t;

typedef struct imsm_vol {
    u_int32_t curr_migr_unit_lo;
    u_int32_t checkpoint_id; /* id to access curr_migr_unit */
    u_int8_t migr_state;     /* Normal or Migrating */
#define MIGR_INIT         0
#define MIGR_REBUILD      1
#define MIGR_VERIFY       2 /* analagous to echo check > sync_action */
#define MIGR_GEN_MIGR     3
#define MIGR_STATE_CHANGE 4
#define MIGR_REPAIR       5
    u_int8_t migr_type; /* Initializing, Rebuilding, ... */
#define RAIDVOL_CLEAN          0
#define RAIDVOL_DIRTY          1
#define RAIDVOL_DSRECORD_VALID 2
    u_int8_t dirty;
    u_int8_t fs_state;       /* fast-sync state for CnG (0xff == disabled) */
    u_int16_t verify_errors; /* number of mismatches */
    u_int16_t bad_blocks;    /* number of bad blocks during verify */
    u_int32_t curr_migr_unit_hi;
    u_int32_t _filler[3];
    struct imsm_map map[1];
    /* here comes another one if migr_state */
} imsm_vol_t;

typedef struct imsm_dev {
    u_int8_t volume[MAX_RAID_SERIAL_LEN];
    u_int32_t size_low;
    u_int32_t size_high;
#define DEV_BOOTABLE            __cpu_to_le32(0x01)
#define DEV_BOOT_DEVICE         __cpu_to_le32(0x02)
#define DEV_READ_COALESCING     __cpu_to_le32(0x04)
#define DEV_WRITE_COALESCING    __cpu_to_le32(0x08)
#define DEV_LAST_SHUTDOWN_DIRTY __cpu_to_le32(0x10)
#define DEV_HIDDEN_AT_BOOT      __cpu_to_le32(0x20)
#define DEV_CURRENTLY_HIDDEN    __cpu_to_le32(0x40)
#define DEV_VERIFY_AND_FIX      __cpu_to_le32(0x80)
#define DEV_MAP_STATE_UNINIT    __cpu_to_le32(0x100)
#define DEV_NO_AUTO_RECOVERY    __cpu_to_le32(0x200)
#define DEV_CLONE_N_GO          __cpu_to_le32(0x400)
#define DEV_CLONE_MAN_SYNC      __cpu_to_le32(0x800)
#define DEV_CNG_MASTER_DISK_NUM __cpu_to_le32(0x1000)
    u_int32_t status;          /* Persistent RaidDev status */
    u_int32_t reserved_blocks; /* Reserved blocks at beginning of volume */
    u_int8_t migr_priority;
    u_int8_t num_sub_vols;
    u_int8_t tid;
    u_int8_t cng_master_disk;
    u_int16_t cache_policy;
    u_int8_t cng_state;
    u_int8_t cng_sub_state;
    u_int16_t my_vol_raid_dev_num; /* Used in Unique volume Id for this RaidDev */

    /* NVM_EN */
    u_int8_t nv_cache_mode;
    u_int8_t nv_cache_flags;

    /* Unique Volume Id of the NvCache Volume associated with this volume */
    u_int32_t nvc_vol_orig_family_num;
    u_int16_t nvc_vol_raid_dev_num;

#define RWH_OFF                            0
#define RWH_DISTRIBUTED                    1
#define RWH_JOURNALING_DRIVE               2
#define RWH_MULTIPLE_DISTRIBUTED           3
#define RWH_MULTIPLE_PPLS_JOURNALING_DRIVE 4
#define RWH_MULTIPLE_OFF                   5
#define RWH_BITMAP                         6
    u_int8_t rwh_policy;                     /* Raid Write Hole Policy */
    u_int8_t jd_serial[MAX_RAID_SERIAL_LEN]; /* Journal Drive serial number */
    u_int8_t _filler1;

#define IMSM_DEV_FILLERS 3
    u_int32_t _filler[IMSM_DEV_FILLERS];
    imsm_vol_t vol;
} imsm_dev_t;

typedef struct imsm_super {
    u_int8_t sig[MAX_SIGNATURE_LENGTH]; /* 0x00 - 0x1F */
    u_int32_t check_sum;                /* 0x20 - 0x23 MPB Checksum */
    u_int32_t mpb_size;                 /* 0x24 - 0x27 Size of MPB */
    u_int32_t family_num;               /* 0x28 - 0x2B Checksum from first time this config was written */
    u_int32_t generation_num;           /* 0x2C - 0x2F Incremented each time this array's MPB is written */
    u_int32_t error_log_size;           /* 0x30 - 0x33 in bytes */
    u_int32_t attributes;               /* 0x34 - 0x37 */
    u_int8_t num_disks;                 /* 0x38 Number of configured disks */
    u_int8_t num_raid_devs;             /* 0x39 Number of configured volumes */
    u_int8_t error_log_pos;             /* 0x3A  */
    u_int8_t fill[1];                   /* 0x3B */
    u_int32_t cache_size;               /* 0x3c - 0x40 in mb */
    u_int32_t orig_family_num;          /* 0x40 - 0x43 original family num */
    u_int32_t pwr_cycle_count;          /* 0x44 - 0x47 simulated power cycle count for array */
    u_int32_t bbm_log_size;             /* 0x48 - 0x4B - size of bad Block Mgmt Log in bytes */
    u_int16_t num_raid_devs_created;    /* 0x4C - 0x4D Used for generating unique
                                         * volume IDs for raid_dev created in this array
                                         * (starts at 1)
                                         */
    u_int16_t _filler1;                 /* 0x4E - 0x4F */
    u_int64_t creation_time;            /* 0x50 - 0x57 Array creation time */
#define IMSM_FILLERS 32
    u_int32_t _filler[IMSM_FILLERS];    /* 0x58 - 0xD7 RAID_MPB_FILLERS */
    struct imsm_disk disk[1];           /* 0xD8 diskTbl[numDisks] */
                                        /* here comes imsm_dev[num_raid_devs] */
                                        /* here comes BBM logs */
} imsm_super_t;

size_t sizeof_imsm_map(imsm_map_t* map);
imsm_map_t* get_imsm_map(imsm_dev_t* dev, int second_map);
size_t sizeof_imsm_dev(imsm_dev_t* dev, int migr_state);
imsm_dev_t* get_imsm_dev(imsm_super_t* mpb, u_int8_t index);

u_int32_t sb_get_effective_family(imsm_super_t* sb);
boolean sb_equal(imsm_super_t* lhs, imsm_super_t* rhs);
sector_count_t get_imsm_dev_size(imsm_dev_t* dev);
sector_count_t get_imsm_map_blocks_per_member(imsm_map_t* map);
sector_count_t get_imsm_map_pba_of_lba0(imsm_map_t* map);

